import os

import click
from pvHelpers.crypto_client import CryptoClient
from pvHelpers.cs_client import BackendClient
from pvHelpers.utils import CaseInsensitiveDict, jdumps, randUnicode

from .utils import add_member, create_export_group, create_new_org, generate_load


@click.command('generate')
@click.option(
    '-u', '--admin-user', 'admin_user',
    required=False,
    default=None,
    type=click.STRING)
@click.option('-c', '--create-org', 'create_org', flag_value=True, default=False)
@click.pass_context
def generate(gctx, admin_user, create_org):
    '''generate mail load on an organization'''

    cs_client = BackendClient()
    cs_client.init(unicode(os.environ['CS_URL']))
    crypto_client = CryptoClient(unicode(os.environ['CRYPTO_URL']))

    if create_org:
        if len(crypto_client.list_local_users()['users']):
            click.echo('Already configured!')

            gctx.exit(1)

        admin_id = u'{}@preveil.test'.format(randUnicode())
        admin = create_new_org(admin_id, crypto_client)

        export_group_size = (int(os.environ.get('EXPORT_GROUP_SIZE', 3)), int(os.environ.get('OPTIONALS_REQUIRED', 1)))
        approvers = []
        for _ in xrange(export_group_size[0]):
            approvers.append(add_member(admin, crypto_client, cs_client))

        group = create_export_group(admin, approvers, export_group_size[1], crypto_client)
        click.echo(jdumps({
            'admin_id': admin['user_id'],
            'org_id': admin['org_info']['org_id'],
            'export_group': group,
            'optionals_required': export_group_size[1],
            'approvers': [{'user_id': a['user_id'], 'account_version': 0} for a in approvers]
        }))

        gctx.exit(0)

    if admin_user is None:
        click.echo('Should provide admin user_id via `-u/--admin-user`!')
        gctx.exit(1)

    generate_load(admin_user, cs_client, crypto_client)


generate()



import time

import pytest

from .compose_helpers import (bootstrap_relay_tool, bootstrap_test_organization,
                              new_test_context, provision_all_services,
                              start_load_generator, stop_all_services, stop_load_generator)
from ..utils import get_rand_port


@pytest.mark.end2end
def end2end_test():
    test_id = new_test_context()
    crypto_port = get_rand_port()
    provision_all_services(test_id, crypto_port)
    organization_info = bootstrap_test_organization(test_id)
    bootstrap_relay_tool(test_id, crypto_port, organization_info)
    c_id = start_load_generator(test_id, organization_info)

    time.sleep(50)
    stop_load_generator(test_id, c_id)

    # TODO: get server_state state

    stop_all_services(test_id)

    # TODO: expect_results(server_state, i)



import os
import random

from mail_relay.config import Config, CSConfig, IMAPConfig, SMTPConfig, VolumeConfig
from mail_relay.store.migrate import migrate
from pvHelpers.cs_client import BackendClient
from pvHelpers.utils import randUnicode
import pytest


@pytest.fixture(scope='session')
def integration_config():
    IMAPConfig(
        os.environ['IMAP_HOST'], int(os.environ['IMAP_PORT']), None, None, False)
    VolumeConfig(os.environ['EXPORT_VOLUME'])
    return Config(
        CSConfig(os.environ['CS_HOST'], int(os.environ['CS_PORT']), False),
        smtp=SMTPConfig(os.environ['SMTP_HOST'], int(os.environ['SMTP_PORT']), None, None, False, None))


@pytest.fixture(scope='session')
def unit_config():
    return Config(
        CSConfig(randUnicode(), random.randint(0, 999999), False),
        smtp=SMTPConfig(randUnicode(), random.randint(0, 999999), None, None, False, None))


@pytest.fixture(scope='session')
def cs_client(integration_config):
    c = BackendClient()
    c.init(integration_config.cs.http)
    return c


@pytest.fixture(scope='module')
def store(tmpdir_factory):
    s = tmpdir_factory.mktemp('store')
    store_path = os.path.abspath(os.path.join(str(s), 'test.sqlite'))
    migrate(store_path)
    return store_path


@pytest.fixture(scope='module')
def get_store(tmpdir_factory):
    def _new_store():
        s = tmpdir_factory.mktemp('store')
        store_path = os.path.abspath(os.path.join(str(s), 'test.sqlite'))
        migrate(store_path)
        return store_path
    return _new_store



from .key import KeyBucket


class TempDeviceKeyBycket(KeyBucket):
    NAME = u'temp_device_key'
    CURRENT_VERSION = 0

    @staticmethod
    def key(user_id, account_version, key_version):
        return '{};{}'.format(account_version, user_id.lower())



import time

import click

from .compose_helpers import (bootstrap_relay_tool, bootstrap_test_organization, get_test_ids,
                              new_test_context, provision_all_services, remove_all_test_contexts,
                              start_load_generator, stop_all_services, stop_load_generator)
from ..utils import get_rand_port


@click.command()
@click.option('-l', '--list-tests', 'list_tests', flag_value=True, default=False)
@click.option('-c', '--clear', 'clear', flag_value=True, default=False)
@click.option(
    '-i', '--test-id', 'test_id',
    type=click.INT, default=None,
    help='Continue specified test number.')
@click.pass_context
def cli(gctx, test_id, list_tests, clear):
    '''handle integration test runner state'''
    if (clear):
        remove_all_test_contexts()
        gctx.exit(0)

    if (list_tests):
        test_ids = get_test_ids()
        click.echo(', '.join([str(i) for i in sorted(test_ids)]))
        gctx.exit(0)

    if test_id is None:
        # create new test context
        click.echo('creating new test context!')
        test_id = new_test_context()
        click.echo('test context created!')

        click.echo('provisioning collection server!')
        click.echo('provisioning crypto server!')
        click.echo('provisioning load generator!')
        crypto_port = get_rand_port()
        provision_all_services(test_id, crypto_port)
        click.echo('crypto server instance ready!')
        click.echo('collection server instance ready!')

        # create new user/org
        click.echo('creating test organization!')
        organization_info = bootstrap_test_organization(test_id)
        click.echo('created test organization!')
        click.echo(organization_info)

        click.echo('configuring relay instance!')
        bootstrap_relay_tool(test_id, crypto_port, organization_info)
        click.echo('relay service started!')

        click.echo('starting load_generator service!')
        c_id = start_load_generator(test_id, organization_info)
        click.echo('load_generator service started!')

        click.echo('waiting for 50 seconds!')  # ~ 60 users , 300 emails (needs to be async)
        time.sleep(50)
        stop_load_generator(test_id, c_id)
        click.echo('load_generator service stopped!')

        # get server_state state

        click.echo('stopping all services!')
        stop_all_services(test_id)

        # expect_results(server_state, i)

    else:
        pass
        # NOTE: this needs collection server container to keep state
        # crypto_port = 0
        # sp_env = MergeDicts(p_env, {
        #     'COMPOSE_PROJECT_NAME': str(test_id),
        #     'CRYPTO_PORT': str(crypto_port)
        # })
        # click.echo('starting all services!')
        # subprocess.check_call(
        #     ['docker-compose', 'up', '-d'],
        #     cwd=TEST_RUNNER_ROOT, env=sp_env)

        # click.echo('starting load_generator service!')
        # p = subprocess.Popen(
        #     ['docker-compose', 'run', '--rm', '-d', 'load_generator', '-u', organization_info['admin_id']],
        #     stdout=subprocess.PIPE, cwd=TEST_RUNNER_ROOT,
        #     env=MergeDicts(sp_env, {'NEW_MEMBER': str(5), 'NEW_EMAIL': str(1)}))
        # c_id = p.stdout.read().strip()
        # assert p.wait() == 0
        # click.echo('load_generator service started!')

        # click.pause('Press any button to stop test run ...')

        # p = subprocess.Popen(
        #     ['docker', 'stop', c_id],
        #     stdout=subprocess.PIPE, cwd=TEST_RUNNER_ROOT, env=sp_env)
        # assert p.stdout.read().strip() == c_id
        # assert p.wait() == 0
        # click.echo('load_generator service stopped!')

        # click.echo('stopping all services!')
        # subprocess.check_call(
        #     ['docker-compose', 'stop'], cwd=TEST_RUNNER_ROOT, env=sp_env)



import time

from pvHelpers.cs_client.utils import ExpiredDeviceKey, ExpiredUserKey
from pvHelpers.logger import g_log
import requests

from .mail_export_scheduler import MailExportScheduler


class StateMachine(object):
    def __init__(self, state):
        self._ctx = state

    def run_forever(self):
        while True:
            try:
                export_state, err = self._ctx.get_set_export_state()
                if err is not None:
                    g_log.error(err)
                    time.sleep(1)
                    continue

                # given valid `export_state` init export scheduler
                scheduler = MailExportScheduler(self._ctx)
                scheduler.run()
            except ExpiredDeviceKey as e:
                # g_log.exception(e.exception)
                g_log.info(u'rotating device key version {} for {}'.format(e.key_version, e.user_id))
                self._ctx.rotate_device_key(e.user_id, e.account_version)
            except ExpiredUserKey as e:
                # if failure for exporter we must terminate process
                pass
            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 401:
                    # TODO: handle intermittent device key rotation
                    pass
                raise
            except Exception as e:
                g_log.exception(e)
                raise
            finally:
                time.sleep(0.5)



from pvHelpers.crypto.utils import CryptoException
from pvHelpers.logger import g_log
from pvHelpers.mail.email import EmailException, PROTOCOL_VERSION
from pvHelpers.utils import b64dec, EncodingException


ACCEPTED_PROTOCOL_VERSIONS = [PROTOCOL_VERSION.V4, PROTOCOL_VERSION.V5, PROTOCOL_VERSION.V6]


def reconstruct_content_blocks(block_ids, blocks, symm_key):
    part = ''
    for block_id in block_ids:
        encrypted_block = blocks.get(block_id)
        if encrypted_block is None:
            return None, KeyError('missing block with block_id {}'.format(block_id))

        encrypted_block = encrypted_block.get('data')
        if encrypted_block is None:
            return None, KeyError('data missing')

        try:
            p = symm_key.decrypt(b64dec(encrypted_block))
        except (CryptoException, EncodingException) as e:
            return None, e
        part += p

    return part, None


def load_content(email, blocks, decrypt_key):
    content_ref_ids = [
        email.body.reference_id] + map(lambda att: att.content.reference_id, email.attachments)

    contents = {}
    for reference_id in content_ref_ids:
        raw_content, err = reconstruct_content_blocks(reference_id.split(','), blocks, decrypt_key)
        if err is not None:
            g_log.exception(err)
            continue
        contents[reference_id] = raw_content

    if not all(map(lambda ref_id: ref_id in contents, content_ref_ids)):
        raise EmailException('Failed to reconstuct email contents, missing blocks')

    body_content = contents[email.body.reference_id]
    email.load_body(body_content)
    for att in email.attachments:
        if att.content.reference_id in contents:
            att.load_content(contents[att.content.reference_id])

    return email


def wrap_for_export(mime):
    # add header
    return mime



from pvHelpers.request.export_local_state import ExportRequestLocalState
from pvHelpers.utils import CaseInsensitiveDict


class ExportState(ExportRequestLocalState):
    def __init__(self, request, export_group, approvers_info):
        super(ExportState, self).__init__(
            request, approvers_info, ExportRequestLocalState.INIT, u'', export_group)

    @property
    def valid_for(self):
        return 60 * 60 * 10

    @property
    def export_group(self):
        return self.group_info

    @property
    def approver_signatures(self):
        return CaseInsensitiveDict(
            {id_: info['approver_signature'] for id_, info in self.approvers_info.iteritems()})

    @property
    def members(self):
        return list(set([u['user_id'] for u in self.request.data['users']]))



import os
import random
import shutil
import subprocess
import time

import click
from pvHelpers.utils import jloads, MergeDicts

from ..utils import crypto_send_user_key

TEST_RUNNER_ROOT = os.path.dirname(os.path.abspath(__file__))
TEST_DIR = os.path.join(TEST_RUNNER_ROOT, '.test_data')
TEST_PREFIX = 'test_'
CONFIGS_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'configs')
P_ENV = os.environ.copy()


def test_dir(test_id):
    return os.path.join(TEST_DIR, '{}{}'.format(TEST_PREFIX, test_id))


def get_test_ids():
    return [int(d.split('_')[1]) for d in os.listdir(TEST_DIR) if d.startswith(TEST_PREFIX)]


def new_test_context():
    existing_test_ids = get_test_ids()
    next_id = max([-1] + existing_test_ids) + 1
    next_dir = test_dir(next_id)
    os.mkdir(next_dir)
    # create volumes
    for v in ['crypto', 'export', 'relay']:
        os.mkdir(os.path.join(next_dir, v))

    # set config file for relay tool
    shutil.copyfile(
        os.path.join(CONFIGS_DIR, 'volume.test.yml'),
        os.path.join(next_dir, 'relay', 'config.yml'))

    return next_id


def remove_all_test_contexts():
    for i in get_test_ids():
        subprocess.check_call(
            ['docker-compose', 'down', '-v'], cwd=TEST_RUNNER_ROOT,
            env=MergeDicts(P_ENV, {'COMPOSE_PROJECT_NAME': str(i), 'CRYPTO_PORT': str(0)}))

        shutil.rmtree(test_dir(i))


def provision_all_services(test_id, crypto_port):
    sp_env = MergeDicts(P_ENV, {
        'COMPOSE_PROJECT_NAME': str(test_id),
        'CRYPTO_PORT': str(crypto_port)
    })
    subprocess.check_call(
        ['docker-compose', 'up', '-d', '--build', '--force-recreate'],
        cwd=TEST_RUNNER_ROOT, env=sp_env)
    # idle some time for services to become reachable
    time.sleep(15)


def stop_all_services(test_id):
    sp_env = MergeDicts(P_ENV, {
        'COMPOSE_PROJECT_NAME': str(test_id),
        'CRYPTO_PORT': str(0)
    })
    subprocess.check_call(
        ['docker-compose', 'stop'], cwd=TEST_RUNNER_ROOT, env=sp_env)


def bootstrap_test_organization(test_id):
    sp_env = MergeDicts(P_ENV, {
        'COMPOSE_PROJECT_NAME': str(test_id),
        'CRYPTO_PORT': str(0)
    })
    p = subprocess.Popen(
        ['docker-compose', 'run', '--rm', 'load_generator', '-c'],
        stdout=subprocess.PIPE, cwd=TEST_RUNNER_ROOT, env=sp_env)
    # print p.stdout.read()
    assert p.wait() == 0
    return jloads(unicode(p.stdout.read()))


def bootstrap_relay_tool(test_id, crypto_port, organization_info):
    sp_env = MergeDicts(P_ENV, {
        'COMPOSE_PROJECT_NAME': str(test_id),
        'CRYPTO_PORT': str(crypto_port)
    })
    click.echo('migrating database!')
    subprocess.check_call(
        ['docker-compose', 'run', '--rm', 'relay', 'migrate'],
        cwd=TEST_RUNNER_ROOT, env=sp_env)
    click.echo('relay database migrated!')

    click.echo('configuring relay instance!')
    subprocess.check_call(
        ['docker-compose', 'run', '--rm', 'relay', 'config', '-c', './data/config.yml'],
        cwd=TEST_RUNNER_ROOT, env=sp_env)
    click.echo('relay configuration saved!')

    click.echo('configuring relay exporter!')
    p = subprocess.Popen(
        ['docker-compose', 'run', '--rm', 'relay', 'config', 'exporter', organization_info['admin_id']],
        stdout=subprocess.PIPE, cwd=TEST_RUNNER_ROOT, env=sp_env)
    send_pin_handle = crypto_send_user_key('127.0.0.1', crypto_port, organization_info['admin_id'])
    time.sleep(3)
    p.stdout.readline()  # skip just info line
    pin = p.stdout.readline().split('pin: ')[1].strip()
    send_pin_handle(pin)
    assert p.wait() == 0
    click.echo('relay exporter {} configured!'.format(organization_info['admin_id']))

    click.echo('configuring approvers!')
    for a in random.sample(organization_info['approvers'], organization_info['optionals_required']):
        p = subprocess.Popen(
            ['docker-compose', 'run', '--rm', 'relay', 'config', 'approver', a['user_id']],
            stdout=subprocess.PIPE, cwd=TEST_RUNNER_ROOT, env=sp_env)
        send_pin_handle = crypto_send_user_key('127.0.0.1', crypto_port, a['user_id'])
        time.sleep(3)
        p.stdout.readline()  # skip just info line
        pin = p.stdout.readline().split('pin: ')[1].strip()
        send_pin_handle(pin)
        assert p.wait() == 0
        click.echo('relay approver {} configured!'.format(a['user_id']))

    click.echo('starting relay service!')
    subprocess.check_call(
        ['docker-compose', 'start', 'relay'],
        stdout=subprocess.PIPE, cwd=TEST_RUNNER_ROOT, env=sp_env)


def start_load_generator(test_id, organization_info):
    sp_env = MergeDicts(P_ENV, {
        'COMPOSE_PROJECT_NAME': str(test_id),
        'CRYPTO_PORT': str(0)
    })
    p = subprocess.Popen(
        ['docker-compose', 'run', '--rm', '-d', 'load_generator', '-u', organization_info['admin_id']],
        stdout=subprocess.PIPE, cwd=TEST_RUNNER_ROOT,
        env=MergeDicts(sp_env, {'NEW_MEMBER': str(5), 'NEW_EMAIL': str(1)}))
    c_id = p.stdout.read().strip()
    assert p.wait() == 0
    return c_id


def stop_load_generator(test_id, c_id):
    sp_env = MergeDicts(P_ENV, {
        'COMPOSE_PROJECT_NAME': str(test_id),
        'CRYPTO_PORT': str(0)
    })
    p = subprocess.Popen(
        ['docker', 'stop', c_id],
        stdout=subprocess.PIPE, cwd=TEST_RUNNER_ROOT, env=sp_env)
    assert p.stdout.read().strip() == c_id
    assert p.wait() == 0



from mail_relay.mail.helpers import ACCEPTED_PROTOCOL_VERSIONS, load_content, wrap_for_export
from pvHelpers.crypto import CryptoException, PVKeyFactory
from pvHelpers.logger import g_log
from pvHelpers.mail import EmailFactory
from pvHelpers.mail.email.helpers import EmailException
from pvHelpers.mail.helpers import decrypt_server_message, get_sender, get_wrapped_key
from pvHelpers.utils import b64dec, CaseInsensitiveDict


class MailExportScheduler():
    def __init__(self, ctx):
        self.ctx = ctx

    def run(self):
        export_state, err = self.ctx.get_set_export_state()
        if err is not None:
            return err

        for member_id in export_state.members:
            member = self.ctx.get_member(member_id, -1)
            err = self.export_user_mails(member, export_state)
            if err is not None:
                g_log.error(err)

    def download_blocks(self, collection_id, block_ids):
        downloaded_server_blocks = CaseInsensitiveDict()
        chunk_size = 50
        for i in range(0, len(block_ids), chunk_size):
            downloaded_server_blocks.update(
                self.ctx.cs_client.get_blocks(self.ctx.exporter, block_ids[i:i+chunk_size], collection_id))

        return downloaded_server_blocks

    def export_user_mails(self, member, export_state):  # noqa: C901
        has_more = True

        while has_more:
            email_updates = self.ctx.cs_client.get_mail_history_for_export(
                self.ctx.exporter, self.ctx.get_members_last_rev_id(member.user_id, member.account_version),
                member.user_id, member.mail_cid, export_state.request.request_id)
            has_more = email_updates['has_more']
            if len(email_updates['messages']) == 0:
                continue

            for m in email_updates['messages']:
                if m['protocol_version'] not in ACCEPTED_PROTOCOL_VERSIONS:
                    g_log.info('unsupported protocol version {} for {} for user ({}, {})'.format(
                        m['protocol_version'], m['id'], member.user_id, member.account_version))
                    continue
                key_version, wrapped_key = get_wrapped_key(m)
                member_encryption_key = self.ctx.get_set_member_encryption_key(
                    member.user_id, key_version)
                if not member_encryption_key:
                    g_log.error(
                        'Decryption key for ({}, {}) not found in'
                        'export req {}.'.format(member.user_id, key_version, export_state.request.request_id))
                    continue

                try:
                    wrapped_raw_key = b64dec(wrapped_key)
                    raw_key = member_encryption_key.unseal(wrapped_raw_key)
                    decrypt_key = PVKeyFactory.deserializeSymmKey(raw_key)

                    decrypted_message = decrypt_server_message(m, member_encryption_key, decrypt_key)

                    if decrypted_message['protocol_version'] <= 4:
                        decrypted_message['collection_id'] = member.mail_cid
                    else:
                        sender_id, key_version = get_sender(decrypted_message)
                        decrypted_message['collection_id'] = self.ctx.cs_client.fetch_users(
                            self.ctx.exporter, [(sender_id, -1)])[sender_id].mail_cid

                    email = EmailFactory.from_server_message(
                        member.user_id, decrypted_message, wrapped_key, key_version, None)

                    blocks = self.download_blocks(
                        decrypted_message['collection_id'],
                        reduce(lambda acc, att: acc + att.content.block_ids, email.attachments, email.body.block_ids))

                    email = load_content(email, blocks, decrypt_key)
                    mime_string = email.to_mime().to_string()
                except CryptoException as e:
                    g_log.error(
                        'error decryption email {} for user ({}, {})'.format(
                            m['id'], member.user_id, member.account_version))

                    g_log.exception(e)
                    continue
                except EmailException as e:
                    g_log.error('error reconstructing email {}'.format(m['id']))
                    g_log.exception(e)
                    continue

                # catch wrapping errors ?
                export_mime = wrap_for_export(mime_string)
                err = self.ctx.relay_mime(
                    member.user_id, member.account_version, m['id'], m['rev_id'], export_mime)
                if err is not None:
                    return 'failed to relay message {} for user ({}, {}): {}'.format(
                        m['id'], member.user_id, member.account_version, err)



# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

setup(
    name='load_generator',
    version='0.0.1',
    author='Saman Tehrani',
    author_email='samanrtehrani@gmail.com',
    entry_points={
      'console_scripts': [
        'generate = load_generator:generate'
      ]
    },
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    python_requires='>=2.7'
)



import click
from mail_relay.daemon.main import main

from .context import ctx


@click.command()
@click.option('--members', 'members', required=False, multiple=True)
@ctx
def start(ctx, members):
    '''Start relay daemon.'''
    from mail_relay.store.handlers import read_exporter

    exporter_a_v, exporter_id = read_exporter(ctx.store.path)
    if not exporter_id:
        click.echo('No exporter configured! Try configuring exporter '
                   '```relay config exporter ${user_id}```, then try again.')
        return

    main(ctx.store.path, members)
